// Squawk messenger. 
// Copyright (C) 2019  Yury Gubich <blue@macaw.me>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef CORE_DATABASE_H
#define CORE_DATABASE_H

#include <map>
#include <string>
#include <optional>

#include <QString>
#include <QStandardPaths>
#include <QDir>
#include <lmdb.h>

class DataBase
{
    class _Table;
    template<class T>
    class Serializer;
public:
    template <class K, class V>
    class Table;
    template <class K, class V>
    class Cache;

    DataBase(const QString& name, uint16_t mapSize = 10);
    ~DataBase();

    void open();
    void close();
    bool ready() const;
    bool removeDirectory();
    QString createDirectory();
    QString getName() const;
    void drop();


    template <class K, class V>
    Table<K, V>* addTable(const std::string& name);

    template <class K, class V>
    Cache<K, V>* addCache(const std::string& name);

    template <class K, class V>
    Table<K, V>* getTable(const std::string& name);

    template <class K, class V>
    Cache<K, V>* getCache(const std::string& name);

public:
    //exceptions
    class Exception;
    class Directory;
    class Closed;
    class Opened;
    class NotFound;
    class Exist;
    class Unknown;

private:
    std::string name;
    bool opened;
    uint16_t size;
    MDB_env* environment;
    std::map<std::string, _Table*> tables;
};

#include "exceptions.h"
#include "operators.hpp"

template <class K, class V>
DataBase::Table<K, V>* DataBase::addTable(const std::string& p_name) {
    if (opened) {
        throw Opened(name, "add table " + p_name);
    }
    DataBase::Table<K, V>* table = new DataBase::Table<K, V>(p_name, this);
    tables.insert(std::make_pair(p_name, (_Table*)table));
    return table;
}

template<class K, class V>
DataBase::Cache<K, V> * DataBase::addCache(const std::string& p_name) {
    if (opened) {
        throw Opened(name, "add cache " + p_name);
    }
    DataBase::Cache<K, V>* cache = new DataBase::Cache<K, V>(p_name, this);
    tables.insert(std::make_pair(p_name, (_Table*)cache));
    return cache;
}

template <class K, class V>
DataBase::Table<K, V>* DataBase::getTable(const std::string& p_name) {
    return static_cast<DataBase::Table<K, V>*>(tables.at(p_name));
}

template <class K, class V>
DataBase::Cache<K, V>* DataBase::getCache(const std::string& p_name) {
    return static_cast<DataBase::Cache<K, V>*>(tables.at(p_name));
}


#endif // CORE_DATABASE_H
